/******************************************************************/
/* igluPinholeCamera.cpp                                          */
/* -----------------------                                        */
/*                                                                */
/* This class is a fairly straightforward pinhole camera class to */
/*    keep camera data and interactions encapsulated.             */
/*                                                                */
/* Chris Wyman (1/14/2013)                                        */
/******************************************************************/

#include "iglu.h"
using namespace iglu;


IGLUPinholeCamera::IGLUPinholeCamera( vec3 eyePos, vec3 viewVec, vec3 upVec, float fovy, float aspectRatio, float zNear, float zFar ) :
	m_zNear( zNear ), m_zFar( zFar ), m_aspectRatio( aspectRatio ), m_fovy( fovy ), 
	m_eyePos( eyePos ), m_viewVec( viewVec.vNormalize() )
{
	m_upVec = m_viewVec.Cross( upVec ).vNormalize().Cross( m_viewVec );
	UpdateMatrices();
	m_moveSpeed = 0.01f*m_zFar;
}

IGLUPinholeCamera::IGLUPinholeCamera( const IGLUPinholeCamera::Ptr fromCam ) :
	m_zNear( fromCam->m_zNear ), m_zFar( fromCam->m_zFar ), m_aspectRatio( fromCam->m_aspectRatio ), m_fovy( fromCam->m_fovy ), 
	m_eyePos( fromCam->m_eyePos ), m_viewVec( fromCam->m_viewVec ), m_upVec( fromCam->m_upVec ),
	m_moveSpeed( fromCam->m_moveSpeed )
{
	UpdateMatrices();
}

void IGLUPinholeCamera::RotateView ( vec3 relativeVec )
{
	m_viewVec = relativeVec.X()*m_rightVec + relativeVec.Y()*m_upVec - relativeVec.Z()*m_viewVec;
	m_upVec = m_viewVec.Cross( m_upVec ).vNormalize().Cross( m_viewVec );
	UpdateMatrices();
}

void IGLUPinholeCamera::SetAspect( float aspectRatio )
{
	m_aspectRatio = aspectRatio;
	UpdateMatrices();
}

void IGLUPinholeCamera::SetNear  ( float zNear )
{
	m_zNear = zNear;
	UpdateMatrices();
}

void IGLUPinholeCamera::SetFar   ( float zFar )
{
	m_zFar      = zFar;
	m_moveSpeed = 0.01f*m_zFar;
	UpdateMatrices();
}

void IGLUPinholeCamera::SetFovy  ( float fovy )
{
	m_fovy = fovy;
	UpdateMatrices();
}

void IGLUPinholeCamera::UpdateMatrices( void )
{
	m_projMatrix = IGLUMatrix4x4::Perspective( m_fovy, m_aspectRatio, m_zNear, m_zFar );
	m_viewMatrix = IGLUMatrix4x4::LookAt( m_eyePos, m_eyePos+m_viewVec, m_upVec );
	m_rightVec = m_viewVec.Cross( m_upVec ).vNormalize();
}



void IGLUPinholeCamera::MoveForward( float scale )
{
	m_eyePos += scale*m_moveSpeed*m_viewVec;
	UpdateMatrices();
}

void IGLUPinholeCamera::MoveBack   ( float scale )
{
	m_eyePos += -scale*m_moveSpeed*m_viewVec;
	UpdateMatrices();
}

void IGLUPinholeCamera::MoveLeft   ( float scale )
{
	m_eyePos += -scale*m_moveSpeed*m_rightVec;
	UpdateMatrices();
}

void IGLUPinholeCamera::MoveRight  ( float scale )
{
	m_eyePos += scale*m_moveSpeed*m_rightVec;
	UpdateMatrices();
}

void IGLUPinholeCamera::MoveUp     ( float scale )
{
	m_eyePos += scale*m_moveSpeed*m_upVec;
	UpdateMatrices();
}

void IGLUPinholeCamera::MoveDown   ( float scale )
{
	m_eyePos += -scale*m_moveSpeed*m_upVec;
	UpdateMatrices();
}
