/***************************************************************************/
/* igluGeomUtils.cpp                                                       */
/* -------------------                                                     */
/*                                                                         */
/* This "class" is simply a collection of drawing utilities, each is a     */
/*     static method.  In some sense, these really belong as part of other */
/*     classes (e.g., the full-screen texture draw could belong to         */
/*     IGLUTexture).  But these are not really integral to the behavior    */
/*     of the associated classes and carry a lot of overhead (e.g.,        */
/*     shaders) that would clutter up other classes.                       */
/*                                                                         */
/* Chris Wyman (10/13/2011)                                                */
/***************************************************************************/


#include "iglu.h"

using namespace iglu;


namespace {  // anonymous namespace to hide globals inside this file

// All planes should be the same size!!
float xyplane[36] = {0, 0,1, 0,0,-1, -1,-1,0,    0, 1,1, 0,0,-1, 1,-1,0,    0, 0,0, 0,0,-1, -1,1,0,    0, 1,0, 0,0,-1, 1,1,0};
float yzplane[36] = {0, 0,1, 1,0,0, 0,-1,-1,     0, 1,1, 1,0,0, 0,-1,1,     0, 0,0, 1,0,0, 0,1,-1,     0, 1,0, 1,0,0, 0,1,1};
float xzplane[36] = {0, 0,1, 0,1,0, -1,0,-1,     0, 1,1, 0,1,0, 1,0,-1,     0, 0,0, 0,1,0, -1,0,1,     0, 1,0, 0,1,0, 1,0,1};

}

IGLUVertexArray *iglu::IGLUGeometry::CreateQuad( uint type, uint matlid )
{
	float *vptr = xyplane;
	if ( type & IGLU_GEOM_XZ_PLANE ) vptr = xzplane;
	if ( type & IGLU_GEOM_YZ_PLANE ) vptr = yzplane;

	IGLUVertexArray::Ptr tmp = new IGLUVertexArray();
	tmp->SetVertexArray( sizeof( xyplane ), vptr, IGLU_STATIC|IGLU_DRAW );
	tmp->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(6*sizeof(float)) );
	if (type & IGLU_GEOM_USE_NORMAL) 
		tmp->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(3*sizeof(float)) );
	if (type & IGLU_GEOM_USE_TEXCOORD) 
		tmp->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(1*sizeof(float)) );

	return tmp;
}

IGLUVertexArray *iglu::IGLUGeometry::CreateXYTiledQuad( uint type, int tiles )
{
	IGLUVertexArray::Ptr tmp = new IGLUVertexArray();

	int sz = tiles*tiles*6;
	float *fptr = (float *)malloc( sz * 9 * sizeof(float) );
	float *vptr = fptr;

	for (int j=0;j<tiles;j++)
		for (int i=0;i<tiles;i++)
		{
			float y0 = float(i)/tiles;
			float y1 = float(i+1)/tiles;
			float x0 = float(j)/tiles;
			float x1 = float(j+1)/tiles;

			float tl[9] = { 0, x0,y0, 0,0,-1, -1+2*x0,1-2*y0,0 };
			float tr[9] = { 0, x1,y0, 0,0,-1, -1+2*x1,1-2*y0,0 };
			float bl[9] = { 0, x0,y1, 0,0,-1, -1+2*x0,1-2*y1,0 };
			float br[9] = { 0, x1,y1, 0,0,-1, -1+2*x1,1-2*y1,0 };

			*vptr++ = tl[0]; *vptr++ = tl[1]; *vptr++ = tl[2]; *vptr++ = tl[3]; *vptr++ = tl[4]; *vptr++ = tl[5]; *vptr++ = tl[6]; *vptr++ = tl[7]; *vptr++ = tl[8];
			*vptr++ = bl[0]; *vptr++ = bl[1]; *vptr++ = bl[2]; *vptr++ = bl[3]; *vptr++ = bl[4]; *vptr++ = bl[5]; *vptr++ = bl[6]; *vptr++ = bl[7]; *vptr++ = bl[8];
			*vptr++ = tr[0]; *vptr++ = tr[1]; *vptr++ = tr[2]; *vptr++ = tr[3]; *vptr++ = tr[4]; *vptr++ = tr[5]; *vptr++ = tr[6]; *vptr++ = tr[7]; *vptr++ = tr[8];
			*vptr++ = bl[0]; *vptr++ = bl[1]; *vptr++ = bl[2]; *vptr++ = bl[3]; *vptr++ = bl[4]; *vptr++ = bl[5]; *vptr++ = bl[6]; *vptr++ = bl[7]; *vptr++ = bl[8];
			*vptr++ = br[0]; *vptr++ = br[1]; *vptr++ = br[2]; *vptr++ = br[3]; *vptr++ = br[4]; *vptr++ = br[5]; *vptr++ = br[6]; *vptr++ = br[7]; *vptr++ = br[8];
			*vptr++ = tr[0]; *vptr++ = tr[1]; *vptr++ = tr[2]; *vptr++ = tr[3]; *vptr++ = tr[4]; *vptr++ = tr[5]; *vptr++ = tr[6]; *vptr++ = tr[7]; *vptr++ = tr[8];
		}

	tmp->SetVertexArray( sz*9*sizeof(float), fptr, IGLU_STATIC|IGLU_DRAW );
	tmp->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(6*sizeof(float)) );
	tmp->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(3*sizeof(float)) );
	tmp->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_FLOAT, 9 * sizeof(float), BUFFER_OFFSET(1*sizeof(float)) );

	free( fptr );

	return tmp;
}