/******************************************************************/
/* igluOBJReader.cpp                                              */
/* -----------------------                                        */
/*                                                                */
/* Defines a reader for OBJ model files and associated materials  */
/*                                                                */
/* Chris Wyman (10/20/2011)                                       */
/******************************************************************/

#include "iglu.h"
#include <float.h>

using namespace iglu;

namespace iglu {
struct IGLUOBJTri
{
	IGLUOBJTri(char *m, char *g, char*o) : mtlName(m), grpName(g), objName(o) {}
	int vIdx[3], nIdx[3], tIdx[3], matlID, objectID;
	char *mtlName, *grpName, *objName;
};
} 

IGLUOBJReader::IGLUOBJReader( char *filename, int params ) :
	IGLUFileParser( filename, IGLUFilePath::MODEL ), IGLUModel(), m_vertArr(0), m_keepCPUArray(false),
	m_hasTexCoords(false), m_hasNormals(false), m_hasVertices(false), m_guessNorms(false), m_shaderID(0),
	m_hasMatlID(true), m_curMatlId(0), m_curObjectId(0), m_hasObjectID(true), m_verboseRead(false),
	m_origMin(FLT_MAX, FLT_MAX, FLT_MAX), m_origMax(FLT_MIN, FLT_MIN, FLT_MIN),
	m_bboxMin(FLT_MAX), m_bboxMax(FLT_MIN), m_tmpVertArray(0), m_tmpVertArraySz(-1), m_baseMtlFileName(0)
{
	// Check the parameters
	m_resize          = params & IGLU_OBJ_UNITIZE ? true : false;
	m_center          = params & IGLU_OBJ_CENTER ? true : false;
	m_compactFormat   = params & IGLU_OBJ_COMPACT_STORAGE ? true : false;
	m_loadMtlFile     = params & IGLU_OBJ_NO_MATERIALS ? false : true;
	m_assignObjects   = params & IGLU_OBJ_NO_OBJECTS ? false : true ;
	m_keepCPUArray    = params & IGLU_OBJ_KEEP_CPU_BUFFER ? true : false;
	m_guessNorms      = params & IGLU_OBJ_GUESS_NORMALS ? true : false;
	m_saveAsBinaryVBO = params & IGLU_OBJ_SAVE_BINARY ? true : false;
	m_verboseRead     = params & IGLU_OBJ_VERBOSE_READ ? true : false;

	if (m_verboseRead) printf("(+) Loading and parsing model '%s'...\n", filename);

	// Create the data structure to interface with OpenGL for drawing this object.
	m_vertArr = new IGLUVertexArray();

	// For clarity (later), we have a pointer to the Read_???_Token() method we will
	//    use when reading a facet.  This will be set when we first see a 'f' line.
	FnParserPtr fnFacetParsePtr = NULL;

	// OK, we have a parser.  Now parse through the file
	IGLUOBJTri *curTri = 0;
	char *linePtr = 0, *mtlFilePtr = 0;
	char keyword[64], vertToken[128];
	char fname[256];
	char *currentObj = 0, *currentGrp = 0, *currentMtl = 0;
	int tmpMatlId, tmpObjId;

	while ( (linePtr = this->ReadNextLine()) != NULL )
	{
		// Each OBJ line starts with a keyword/keyletter
		this->GetLowerCaseToken( keyword );
		
		switch( keyword[0] )
		{
		case 'v': 
			if (keyword[1] == 'n')        // We found a normal!
				m_objNorms.Add( this->GetVec3() );
			else if (keyword[1] == 't')   // We found a texture coordinate!
				m_objTexCoords.Add( this->GetVec2() );
			else if (keyword[1] == 0 )    // We found a vertex!
			{
				uint idx = m_objVerts.Add( this->GetVec3() );  
				m_origMax = Max( m_origMax, m_objVerts[idx] );
				m_origMin = Min( m_origMin, m_objVerts[idx] );
			}
			break;
		case 'm': // We found the name of a material file!
			this->GetToken( fname );
			mtlFilePtr = (char *)malloc( strlen(fname)+strlen(fileDirectory)+1 );
			sprintf( mtlFilePtr, "%s%s", fileDirectory, fname );
			m_objMtlFiles.Add( mtlFilePtr ); 
			if (!m_baseMtlFileName)
				m_baseMtlFileName = strdup( fname );
			if (m_loadMtlFile) { 
				// Should be able to do as a local...  but not inside switch?  Do it the hard way.
				delete( new IGLUOBJMaterialReader( mtlFilePtr ));  // Load it.
			}
			break;
		case 'o': // We found a name for the object following this flag
			this->GetToken( fname );
			currentObj = strdup( fname );
			
			//Check if we have already found this object name
			tmpObjId = GetObjectID(currentObj);
			if(-1 != tmpObjId){
				m_curObjectId = uint(tmpObjId);
			}else{ //It is a new object. Add it to the list. 
				m_objObjectNames.Add(currentObj);
				m_curObjectId = m_objObjectNames.Size() - 1;
			}
			
			break;
		case 'g': // We found a name for the group following this flag
			this->GetToken( fname );
			currentGrp = strdup( fname );
			break;
		case 'u': // We found the name of the material we'll be using
			this->GetToken( fname );
			currentMtl = strdup( fname );
			tmpMatlId = IGLUOBJMaterialReader::GetNamedMaterialId( fname );
			if (tmpMatlId >= 0)
				m_curMatlId = uint(tmpMatlId);
			break;
		case 's':
			// There's a smoothing command.  We're ignoring these.
			break;
		case 'f': // We found a facet!
			if (keyword[1] != 0)  // There's some odd models out there with 'fo' or other
				break;            //    additional facet types.  Mostly these are junk!
			curTri = new IGLUOBJTri( currentMtl, currentGrp, currentObj);
			curTri->matlID = m_curMatlId;
			curTri->objectID = m_curObjectId;
			
			//Ensure the 'f' line has at least three entries. 
			if(!IsValidFLine(&fnFacetParsePtr)){
				continue;
			}


			// There are multiple different formats for 'f' lines.  Decide which it is,
			//    and select the appropriate class member we'll use to read this line.
			SelectReadMethod( &fnFacetParsePtr );
		

			// Read first three set of indices on this line
			((*this).*(fnFacetParsePtr))( curTri, 0, NULL );
			((*this).*(fnFacetParsePtr))( curTri, 1, NULL );
			((*this).*(fnFacetParsePtr))( curTri, 2, NULL );
			m_objTris.Add( curTri );

			// Do we have more vertices in this facet?  Check to see if the 
			//    next token is non-empty (if empty the [0]'th char == 0)
			//    If we do, we'll have to triangulate the facet, so make a new tri
			this->GetToken( vertToken );
			while ( vertToken[0] ) 
			{
				curTri = new IGLUOBJTri( currentMtl, currentGrp, currentObj);
				curTri->matlID = m_curMatlId;
				curTri->objectID = m_curObjectId;
				CopyForTriangleFan( curTri );
				((*this).*(fnFacetParsePtr))( curTri, 2, vertToken );
				m_objTris.Add( curTri );
				this->GetToken( vertToken );
			}
			break;

		default:  // We have no clue what to do with this line....
			this->WarningMessage("Found corrupt line in OBJ.  Unknown keyword '%s'", keyword);
		}
	}

	// No need to keep the file hanging around open.
	CloseFile();

	// Initially the current and original bboxes are the same
	m_bboxMin = m_origMin;
	m_bboxMax = m_origMax;

	// If we already have surface normals, there's no need to use facet normals
	if (HasNormals()) m_guessNorms = false;

	// Create the GPU buffers for this object, so we have them laying around later.
	if (m_compactFormat)
	{
		if (m_verboseRead) printf("  (-) Using compact vertex array format...\n");
		GetCompactArrayBuffer();
	}
	else
	{
		if (m_verboseRead) printf("  (-) Using non-compact vertex array format...\n");
		GetArrayBuffer();
		GetElementArrayBuffer();
	}

	// If we asked to save this as a binary object array, save that out.
	if (m_saveAsBinaryVBO)
	{
		if (m_verboseRead) printf("  (-) Saving '%sar' from vertex array...\n", filename);
		char binName[256];
		sprintf( binName, "%sar", filename );
		SaveBinaryVBO( binName );
	}

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  {
		if (m_verboseRead) printf("  (-) Enabling normals for use in shader...\n");
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_normOff));
	}
	if (HasTexCoords()) {
		if (m_verboseRead) printf("  (-) Enabling texture coordinates for use in shader...\n");
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_texOff));
	}
	if (HasMatlID() && m_loadMtlFile) {
		if (m_verboseRead) printf("  (-) Enabling material IDs for use in shader...\n");
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	}
	if (HasObjectID()) {
		if (m_verboseRead) printf("  (-) Enabling object IDs for use in shader...\n");
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_objectIdOff));
	}

	if (m_verboseRead) printf("  (-) Model reading completed successfully! (%d verts, %d tris)\n", m_objVerts.Size(), m_objTris.Size() );
}

IGLUOBJReader::~IGLUOBJReader()
{
	// Free strings contining the mtllib filenames
	for (uint i=0; i<m_objMtlFiles.Size(); i++)
		free( m_objMtlFiles[i] );

	// Get rid of our vertex array
	delete m_vertArr;
}

unsigned int IGLUOBJReader::GetVertexIndex( int relativeIdx )
{
	if (relativeIdx == 0)
	{
		this->WarningMessage("Unexpected OBJ vertex index of 0!");
		relativeIdx = 1;  // 0 is undefined, and will crash.  Give at least a non-segfaultable value;
	}
	return relativeIdx > 0 ? relativeIdx-1 : m_objVerts.Size()+relativeIdx;
}

unsigned int IGLUOBJReader::GetNormalIndex( int relativeIdx )
{
	if (relativeIdx == 0)
		this->WarningMessage("Unexpected OBJ normal index of 0!");
	// Note the >=, which is different than the > in the vertex index.  An index
	//    of 0 is invalid.  Our code treats a -1 return value as "no normal".  That' fine
	//    for nomrals and texture coords
	return relativeIdx >= 0 ? relativeIdx-1 : m_objNorms.Size()+relativeIdx;
}

unsigned int IGLUOBJReader::GetTextureIndex( int relativeIdx )
{
	//if (relativeIdx == 0)
	//	this->WarningMessage("Unexpected OBJ texture coord index of 0!");

	// Note the >=, which is different than the > in the vertex index.  An index
	//    of 0 is invalid.  Our code treats a -1 return value as "no normal".  That' fine
	//    for nomrals and texture coords
	return relativeIdx >= 0 ? relativeIdx-1 : m_objTexCoords.Size()+relativeIdx;
}

int IGLUOBJReader::GetObjectID( char *objName )
{
	//Search the array of objects for one with the same name.
	//If found return the index to that object
	//Otherwise add the new name to the list

	for(uint i = 0; i < m_objObjectNames.Size(); i++){
		if (!strcmp(objName, m_objObjectNames[i])) 
			return int(i);
	} 

	return -1;

}
void IGLUOBJReader::Read_V_Token( IGLUOBJTri *tri, int idx, char *token )
{
	char vertToken[128], *tPtr = vertToken;
	int vIdx;

	// If the user didn't pass in a token, read one
	if (!token)
		this->GetToken( vertToken );
	else 
		tPtr = token;

	// Parse a vertex & normal from the token
	sscanf( tPtr, "%d", &vIdx );

	// Resolve these into indicies in our data structure (not the OBJ file)
	tri->vIdx[idx] = GetVertexIndex( vIdx );
	tri->nIdx[idx] = -1;
	tri->tIdx[idx] = -1;
}

void IGLUOBJReader::Read_VN_Token( IGLUOBJTri *tri, int idx, char *token )
{
	char vertToken[128], *tPtr = vertToken;
	int vIdx, nIdx;

	// If the user didn't pass in a token, read one
	if (!token)
		this->GetToken( vertToken );
	else 
		tPtr = token;

	// Parse a vertex & normal from the token
	sscanf( tPtr, "%d//%d", &vIdx, &nIdx );

	// Resolve these into indicies in our data structure (not the OBJ file)
	tri->vIdx[idx] = GetVertexIndex( vIdx );
	tri->nIdx[idx] = GetNormalIndex( nIdx );
	tri->tIdx[idx] = -1;
}

void IGLUOBJReader::Read_VT_Token( IGLUOBJTri *tri, int idx, char *token )
{
	char vertToken[128], *tPtr = vertToken;
	int vIdx, tIdx;

	// If the user didn't pass in a token, read one
	if (!token)
		this->GetToken( vertToken );
	else 
		tPtr = token;

	// Parse a vertex & normal from the token
	sscanf( tPtr, "%d/%d", &vIdx, &tIdx );

	// Resolve these into indicies in our data structure (not the OBJ file)
	tri->vIdx[idx] = GetVertexIndex( vIdx );
	tri->nIdx[idx] = -1;
	tri->tIdx[idx] = GetTextureIndex( tIdx );
}

void IGLUOBJReader::Read_VTN_Token( IGLUOBJTri *tri, int idx, char *token )
{
	char vertToken[128], *tPtr = vertToken;
	int vIdx, tIdx, nIdx;

	// If the user didn't pass in a token, read one
	if (!token)
		this->GetToken( vertToken );
	else 
		tPtr = token;

	// Parse a vertex & normal from the token
	sscanf( tPtr, "%d/%d/%d", &vIdx, &tIdx, &nIdx );

	// Resolve these into indicies in our data structure (not the OBJ file)
	tri->vIdx[idx] = GetVertexIndex( vIdx );
	tri->nIdx[idx] = GetNormalIndex( nIdx );
	tri->tIdx[idx] = GetTextureIndex( tIdx );
}

void IGLUOBJReader::CopyForTriangleFan( IGLUOBJTri *newTri )
{
	newTri->vIdx[0] = m_objTris[m_objTris.Size()-1]->vIdx[0]; 
	newTri->nIdx[0] = m_objTris[m_objTris.Size()-1]->nIdx[0];
	newTri->tIdx[0] = m_objTris[m_objTris.Size()-1]->tIdx[0];
	newTri->vIdx[1] = m_objTris[m_objTris.Size()-1]->vIdx[2];
	newTri->nIdx[1] = m_objTris[m_objTris.Size()-1]->nIdx[2];
	newTri->tIdx[1] = m_objTris[m_objTris.Size()-1]->tIdx[2];
}

void IGLUOBJReader::SelectReadMethod( FnParserPtr *pPtr )
{
	int v, t, n; // garbage vars
	char vertToken[128];
	char *lPtr = this->GetCurrentLinePointer();

	// Non-destructively parse the next token
	StripLeadingTokenToBuffer( lPtr, vertToken );

	if (strstr(vertToken, "//"))                               // Then it has the v//n format
	{
		*pPtr = &iglu::IGLUOBJReader::Read_VN_Token;
		m_hasVertices = m_hasNormals = true;
	}

	else if ( sscanf(vertToken, "%d/%d/%d", &v, &t, &n) == 3 ) // Then it has the v/t/n format
	{
		*pPtr = &iglu::IGLUOBJReader::Read_VTN_Token;
		m_hasVertices = m_hasNormals = m_hasTexCoords = true;
	}

	else if ( sscanf(vertToken, "%d/%d", &v, &t) == 2 )        // Then it has the v/t format
	{
		*pPtr = &iglu::IGLUOBJReader::Read_VT_Token;
		m_hasVertices = m_hasTexCoords = true;
	}else                                                     // Then it has the v format
		*pPtr = &iglu::IGLUOBJReader::Read_V_Token;
		m_hasVertices = true;
	}
	
bool IGLUOBJReader::IsValidFLine(FnParserPtr *pPtr)
{

	//Garbage variables. This will break if it comes accross an entry on a line
	//longer than 128 characters. 
	char c0[128];
	char c1[128];
	char c2[128];
	char vertToken[128]; 

	// Non-destructively parse the next token
	char *lPtr = this->GetCurrentLinePointer();
	StripLeadingTokenToBuffer( lPtr, vertToken );

	//Make sure we have at least three entries in this vertex
	if(sscanf(lPtr, "%s %s %s", c0, c1, c2) >= 3){
		return true;
	}else{
		this->WarningMessage("Corrupt 'f': %s", lPtr);
		return false; 
	}
}

void IGLUOBJReader::AddDataToArray( float *arr, int startIdx, int matlID, int objectID, vec3 *vert, vec3 *norm, vec2 *tex )
{
	// We'd better have a vertex, or this whole rendering process is stupid....
	assert( vert );

	
	int i = startIdx;
	
	// Add the material ID
	arr[i++] = m_loadMtlFile ? float(matlID) : 0.0f;
	
	// Add the object ID
	arr[i++] = m_assignObjects ? float(objectID) : 0.0f;

	// Add the vertex
	arr[i++] = vert->X();
	arr[i++] = vert->Y();
	arr[i++] = vert->Z();

	// If this vertex has a normal, add it.
	if (norm)
	{
		arr[i++] = norm->X();
		arr[i++] = norm->Y();
		arr[i++] = norm->Z();
	}

	// If this vertex has a texture coordinate, add it.
	if (tex)
	{
		arr[i++] = tex->X();
		arr[i++] = tex->Y();
	}
}

void IGLUOBJReader::CenterAndResize( float *arr, int numVerts )
{
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

	int startIdx   = m_vertOff / sizeof(float);
	int vertStride = m_vertStride / sizeof(float);
	float minX = FLT_MAX, minY = FLT_MAX, minZ = FLT_MAX;
	float maxX = FLT_MIN, maxY = FLT_MIN, maxZ = FLT_MIN;

	// Walk through vertex array, find max & min (x, y, z)
	for (int i=0; i<numVerts; i++)
	{
		int idx = startIdx + i*vertStride;
		minX = iglu::min( minX, arr[idx+0] );
		maxX = iglu::max( maxX, arr[idx+0] );
		minY = iglu::min( minY, arr[idx+1] );
		maxY = iglu::max( maxY, arr[idx+1] );
		minZ = iglu::min( minZ, arr[idx+2] );
		maxZ = iglu::max( maxZ, arr[idx+2] );
	}

	// Find the center of the object & the width of the object
	float ctrX = (m_resize || m_center ? 0.5f*(minX+maxX) : 0.0f );
	float ctrY = (m_resize || m_center ? 0.5f*(minY+maxY) : 0.0f ); 
	float ctrZ = (m_resize || m_center ? 0.5f*(minZ+maxZ) : 0.0f );
	float deltaX = 0.5f*(maxX-minX), deltaY = 0.5f*(maxY-minY), deltaZ = 0.5f*(maxZ-minZ);
	float delta = ( m_resize ? 
		            iglu::max( deltaX, iglu::max( deltaY, deltaZ ) ) :
	                1.0f );

	// Walk through the vertex array and update the positions
	for (int i=0; i<numVerts; i++)
	{
		int idx = startIdx + i*vertStride;
		arr[idx+0] = (arr[idx+0] - ctrX ) / delta;
		arr[idx+1] = (arr[idx+1] - ctrY ) / delta;
		arr[idx+2] = (arr[idx+2] - ctrZ ) / delta;
	}

	// OK, the bounds have changed; update our bounding box
	m_bboxMin = (m_bboxMin - vec3(ctrX, ctrY, ctrZ)) / delta;
	m_bboxMax = (m_bboxMax - vec3(ctrX, ctrY, ctrZ)) / delta;
}

void IGLUOBJReader::GetCompactArrayBuffer( void )
{
	// Create an OBJ vert ID -> element array vert ID map.  Init all entries to 0xFFFFFFFF.
	//    Also, create mapping vertID -> last normal ID used for this vertex
	//          create mapping vertID -> last tex coord used for this vertex
	uint *vertMapping = (uint *)malloc( m_objVerts.Size() * sizeof( uint ) );
	uint *normMapping = (uint *)malloc( m_objVerts.Size() * sizeof( uint ) );
	uint *texMapping  = (uint *)malloc( m_objVerts.Size() * sizeof( uint ) );
	assert( vertMapping );
	memset( vertMapping, 0xFF, m_objVerts.Size() * sizeof( uint ) );
	memset( normMapping, 0xFF, m_objVerts.Size() * sizeof( uint ) );
	memset( texMapping, 0xFF, m_objVerts.Size() * sizeof( uint ) );

	//   We'll have 1 float for a material ID
	//   We'll have 3 floats (x,y,z) for each of the 3 verts of each triangle 
	//   We'll have 3 floats (x,y,z) for each of the 3 normals of each triangle
	//   We'll have 2 floats (u,v) for each of the 3 texture coordinates of each triangle
	uint  numComponents = 1 + 1 + 3 + (m_hasNormals||m_guessNorms ? 3 : 0) + (m_hasTexCoords ? 2 : 0);

	int bufSz  = numComponents * sizeof( float ) * (3 * m_objTris.Size());
	m_vertStride = numComponents * sizeof( float );
	m_matlIdOff  = 0 * sizeof( float );
	m_objectIdOff = 1 * sizeof(float); 
	m_vertOff    = 2 * sizeof( float );
	m_normOff    = (m_hasNormals||m_guessNorms? 5 : 0) * sizeof( float );
	m_texOff     = (m_hasTexCoords ? (m_hasNormals||m_guessNorms? 8 : 5) : 0) * sizeof( float );
	if (m_verboseRead) printf("    (*) Stride: %d, Offsets: v %d, m %d, o %d, n %d t %d\n", m_vertStride, m_vertOff, m_matlIdOff, m_objectIdOff, m_normOff, m_texOff);

	// Create a vertex buffer for us to fill up.
	float *tmpBuf = (float *)malloc( bufSz );
	if (!tmpBuf)
		ErrorExit("Memory allocation error", __FILE__, __FUNCTION__, __LINE__);

	// Create an element array buffer to fill up
	uint elembufSz = 3 * sizeof( unsigned int ) * m_objTris.Size();
	uint *tmpElemBuf = (uint *)malloc( elembufSz );

	// We'll need to know the size of our two arrays
	int numArrayElements = 3 * m_objTris.Size();  // Known in advance
	int numArrayVerts    = 0;                     // Depends on how many verts are reused.  We'll compute

	// A location to store normal guesses
	vec3 tmpNorm;
	vec3 *normGuess = (m_guessNorms ? &tmpNorm : 0);


	for (uint i=0, triNum=0; triNum < m_objTris.Size(); i+=3,triNum++ )
	{
		int i0 = m_objTris[triNum]->vIdx[0];
		int i1 = m_objTris[triNum]->vIdx[1];
		int i2 = m_objTris[triNum]->vIdx[2];

		int *nIdx = m_objTris[triNum]->nIdx;
		int *tIdx = m_objTris[triNum]->tIdx;

		if ( i0 < 0 || i0 > int(m_objVerts.Size()) ) continue;
		if ( i1 < 0 || i1 > int(m_objVerts.Size()) ) continue;
		if ( i2 < 0 || i2 > int(m_objVerts.Size()) ) continue;

		tmpNorm = ((m_objVerts[i1]-m_objVerts[i0]).Cross(m_objVerts[i2]-m_objVerts[i0])).vNormalize();

		if (    (vertMapping[i0] == 0xFFFFFFFF) // We haven't seen this vertex yet.  Add to list
		     || (m_hasNormals && normMapping[i0] != m_objTris[triNum]->nIdx[0])     // We saw this vertex...  but w/different normal
			 || (m_hasTexCoords && texMapping[i0] != m_objTris[triNum]->tIdx[0]) )  // We saw this vertex...  but w/different texcoord
		{
			AddDataToArray( tmpBuf, numArrayVerts*numComponents, m_objTris[triNum]->matlID, 
							m_objTris[triNum]->objectID,
							&m_objVerts[m_objTris[triNum]->vIdx[0]], 
							(m_hasNormals && (nIdx[0]>0)) ? &m_objNorms[nIdx[0]] : normGuess, 
							(m_hasTexCoords && (tIdx[0]>0)) ? &m_objTexCoords[tIdx[0]] : 0 );
			tmpElemBuf[i]   = numArrayVerts;
			vertMapping[i0] = numArrayVerts++;
			normMapping[i0] = nIdx[0];
			texMapping[i0]  = tIdx[0];
		}
		else                               // We've already seen vertex; reuse it.
			tmpElemBuf[i] = vertMapping[i0];

		if (    (vertMapping[i1] == 0xFFFFFFFF) // We haven't seen this vertex yet.  Add to list
		     || (m_hasNormals && normMapping[i1] != m_objTris[triNum]->nIdx[1])     // We saw this vertex...  but w/different normal
			 || (m_hasTexCoords && texMapping[i1] != m_objTris[triNum]->tIdx[1]) )  // We saw this vertex...  but w/different texcoord
		{
			AddDataToArray( tmpBuf, numArrayVerts*numComponents, m_objTris[triNum]->matlID, 
							m_objTris[triNum]->objectID,
							&m_objVerts[m_objTris[triNum]->vIdx[1]], 
							(m_hasNormals && (nIdx[1]>0)) ? &m_objNorms[nIdx[1]] : normGuess, 
							(m_hasTexCoords && (tIdx[1]>0)) ? &m_objTexCoords[tIdx[1]] : 0 );
			tmpElemBuf[i+1]   = numArrayVerts;
			vertMapping[i1] = numArrayVerts++;
			normMapping[i1] = nIdx[1];
			texMapping[i1]  = tIdx[1];
		}
		else                               // We've already seen vertex; reuse it.
			tmpElemBuf[i+1] = vertMapping[i1];

		if (    (vertMapping[i2] == 0xFFFFFFFF) // We haven't seen this vertex yet.  Add to list
		     || (m_hasNormals && normMapping[i2] != m_objTris[triNum]->nIdx[2])     // We saw this vertex...  but w/different normal
			 || (m_hasTexCoords && texMapping[i2] != m_objTris[triNum]->tIdx[2]) )  // We saw this vertex...  but w/different texcoord
		{
			AddDataToArray( tmpBuf, numArrayVerts*numComponents, m_objTris[triNum]->matlID, 
							m_objTris[triNum]->objectID,
							&m_objVerts[m_objTris[triNum]->vIdx[2]], 
							(m_hasNormals && (nIdx[2]>0)) ? &m_objNorms[nIdx[2]] : normGuess, 
							(m_hasTexCoords && (tIdx[2]>0)) ? &m_objTexCoords[tIdx[2]] : 0 );
			tmpElemBuf[i+2]   = numArrayVerts;
			vertMapping[i2] = numArrayVerts++;
			normMapping[i2] = nIdx[2];
			texMapping[i2]  = tIdx[2];
		}
		else                               // We've already seen vertex; reuse it.
			tmpElemBuf[i+2] = vertMapping[i2];
	}

	// If the user asked us to resize & center the object, do that.
	if (m_resize || m_center)
		CenterAndResize( tmpBuf, numArrayVerts );

	// Copy our arrays into their GPU buffers
	m_vertArr->SetVertexArray( numArrayVerts * m_vertStride, tmpBuf, IGLU_STATIC|IGLU_DRAW );
	m_vertArr->SetElementArray( GL_UNSIGNED_INT, elembufSz, tmpElemBuf, IGLU_STATIC|IGLU_DRAW );

	// Free our temporary copy of the data
	if (!m_keepCPUArray)
		free( tmpBuf );
	else
	{
		m_tmpVertArray = tmpBuf;
		m_tmpVertArraySz = numComponents * (3 * m_objTris.Size());
	}
	free( tmpElemBuf );
	free( vertMapping );

	// If we asked to guess normals, we've done it.  Treat everything hereon as if we had norms:
	if (m_guessNorms) m_hasNormals = true;
}

typedef struct {
	uint vboVersionID;       // Magic number / file version header
	uint numVerts;           // Number of vertices in the VBO
	uint numElems;           // Number of elements (i.e., indices in the VBO)
	uint elemType;           // E.g. GL_TRIANGLES
	uint vertStride;         // Number of bytes between subsequent vertices
	uint vertBitfield;       // Binary bitfield describing valid vertex components 
	uint matlOffset;         // In vertex, offset to material ID
	uint objOffset;          // In vertex, offset to object ID
	uint vertOffset;         // In vertex, offset to vertex 
	uint normOffset;         // In vertex, offset to normal 
	uint texOffset;          // In vertex, offset to texture coordinate
	char matlFileName[84];   // Filename containing material information
} IGLUBinaryVBOHeader;

enum {
	HAS_VERTEX   = 0x10,
	HAS_NORMAL   = 0x01,
	HAS_TEXCOORD = 0x02,
	HAS_MATLID   = 0x04,
	HAS_OBJID    = 0x08
} vertBits;

bool IGLUOBJReader::SaveBinaryVBO( char *outFile )
{
	// Make sure we have data before saving.
	GLsizeiptr vertSize = m_vertArr->GetVertexBuffer()->GetBufferSize();
	GLsizeiptr elemSize = m_vertArr->GetElementBuffer()->GetBufferSize();
	if (vertSize <= 0 || elemSize <= 0) return false;

	IGLUBinaryVBOHeader hdr;
	hdr.vboVersionID = 1;
	hdr.numVerts     = vertSize / m_vertStride;
	hdr.numElems     = 3 * m_objTris.Size();
	hdr.elemType     = GL_TRIANGLES;
	hdr.vertStride   = m_vertStride;
	hdr.vertBitfield = (HasVertices()?HAS_VERTEX:0) |
		               (HasNormals()?HAS_NORMAL:0) |
					   (HasTexCoords()?HAS_TEXCOORD:0) |
					   (HasMatlID()?HAS_MATLID:0) |
					   (HasObjectID()?HAS_OBJID:0);
	hdr.matlOffset   = m_matlIdOff;
	hdr.objOffset    = m_objectIdOff;
	hdr.vertOffset   = m_vertOff;
	hdr.normOffset   = m_normOff;
	hdr.texOffset    = m_texOff;
	if (m_baseMtlFileName)
		strncpy( hdr.matlFileName, m_baseMtlFileName, 84 );
	else
		memset( hdr.matlFileName, 0, 84 );

	FILE *f = fopen( outFile, "wb" );
	fwrite( &hdr, sizeof(IGLUBinaryVBOHeader), 1, f );
	void *ptr = m_vertArr->MapVertexArray();
	fwrite( ptr, vertSize, 1, f );
	m_vertArr->UnmapVertexArray();
	ptr = m_vertArr->MapElementArray();
	fwrite( ptr, elemSize, 1, f );
	m_vertArr->UnmapElementArray();
	fclose( f );

	return true;
}

void IGLUOBJReader::GetArrayBuffer( void )
{
	// Using our EXTREMELY naive approach...
	//   We'll have 1 float for a material ID
	//   We'll have 1 float for an object ID
	//   We'll have 3 floats (x,y,z) for each of the 3 verts of each triangle 
	//   We'll have 3 floats (x,y,z) for each of the 3 normals of each triangle
	//   We'll have 2 floats (u,v) for each of the 3 texture coordinates of each triangle
	uint  numComponents = 1 + 1 + 3 + (m_hasNormals||m_guessNorms ? 3 : 0) + (m_hasTexCoords ? 2 : 0);

	int bufSz  = numComponents * sizeof( float ) * (3 * m_objTris.Size());
	m_vertStride = numComponents * sizeof( float );
	m_matlIdOff  = 0 * sizeof( float );
	m_objectIdOff = 1 * sizeof(float); 
	m_vertOff    = 2 * sizeof( float );
	m_normOff    = (m_hasNormals||m_guessNorms ? 5 : 0) * sizeof( float );
	m_texOff     = (m_hasTexCoords ? (m_hasNormals||m_guessNorms ? 8 : 5) : 0) * sizeof( float );

	// Create a vertex array buffer for us to fill up.
	float *tmpBuf = (float *)malloc( bufSz );

	// A location to store normal guesses
	vec3 tmpNorm;
	vec3 *normGuess = (m_guessNorms ? &tmpNorm : 0);

	// For our very, very early reader, we'll use the MOST NAIVE approach
	for (uint i=0, triNum=0; i<3*numComponents*m_objTris.Size(); i+=3*numComponents,triNum++ )
	{
		if ( m_objTris[triNum]->vIdx[0] < 0 || m_objTris[triNum]->vIdx[0] > int(m_objVerts.Size()) ) continue;
		if ( m_objTris[triNum]->vIdx[1] < 0 || m_objTris[triNum]->vIdx[1] > int(m_objVerts.Size()) ) continue;
		if ( m_objTris[triNum]->vIdx[2] < 0 || m_objTris[triNum]->vIdx[2] > int(m_objVerts.Size()) ) continue;

		tmpNorm = ((m_objVerts[m_objTris[triNum]->vIdx[1]]-m_objVerts[m_objTris[triNum]->vIdx[0]]).Cross(m_objVerts[m_objTris[triNum]->vIdx[2]]-m_objVerts[m_objTris[triNum]->vIdx[0]])).vNormalize();

		AddDataToArray( tmpBuf, i, m_objTris[triNum]->matlID,
						m_objTris[triNum]->objectID,
			            &m_objVerts[m_objTris[triNum]->vIdx[0]], 
						m_hasNormals ? &m_objNorms[m_objTris[triNum]->nIdx[0]] : normGuess, 
						m_hasTexCoords ? &m_objTexCoords[m_objTris[triNum]->tIdx[0]] : 0 );
		
		AddDataToArray( tmpBuf, i+numComponents, m_objTris[triNum]->matlID,
						m_objTris[triNum]->objectID,
			            &m_objVerts[m_objTris[triNum]->vIdx[1]], 
						m_hasNormals ? &m_objNorms[m_objTris[triNum]->nIdx[1]] : normGuess, 
						m_hasTexCoords ? &m_objTexCoords[m_objTris[triNum]->tIdx[1]] : 0 );
		
		AddDataToArray( tmpBuf, i+2*numComponents, m_objTris[triNum]->matlID,
						m_objTris[triNum]->objectID,
			            &m_objVerts[m_objTris[triNum]->vIdx[2]], 
						m_hasNormals ? &m_objNorms[m_objTris[triNum]->nIdx[2]] : normGuess, 
						m_hasTexCoords ? &m_objTexCoords[m_objTris[triNum]->tIdx[2]] : 0 );
	}

	// If the user asked us to resize & center the object, do that.
	if (m_resize || m_center)
		CenterAndResize( tmpBuf, 3 * m_objTris.Size() );

	// Copy our element array into the buffer
	m_vertArr->SetVertexArray( bufSz, tmpBuf, IGLU_STATIC|IGLU_DRAW );

	// Free our temporary copy of the data
	if (!m_keepCPUArray)
		free( tmpBuf );
	else
	{
		m_tmpVertArray = tmpBuf;
		m_tmpVertArraySz = numComponents * (3 * m_objTris.Size());
	}

	// If we asked to guess normals, we've done it.  Treat everything hereon as if we had norms:
	if (m_guessNorms) m_hasNormals = true;
}

uint IGLUOBJReader::GetArrayBufferStride( void )
{
	return m_vertStride;
}

void IGLUOBJReader::GetElementArrayBuffer( void )
{
	// We'll have one index for each of the 3 verts of each triangle (i.e., GL_TRIANGLES)
	uint bufSz = 3 * sizeof( unsigned int ) * m_objTris.Size();

	// Create a buffer for us to fill up.
	uint *tmpBuf = (uint *)malloc( bufSz );

	// For our very, very early reader, we'll use the MOST NAIVE approach
	for (uint i=0; i<3*m_objTris.Size(); i++)
		tmpBuf[i] = i;

	// Copy our element array into the buffer
	m_vertArr->SetElementArray( GL_UNSIGNED_INT, bufSz, tmpBuf, IGLU_STATIC|IGLU_DRAW );

	// Free our temporary copy of the data
	free( tmpBuf );
}

int IGLUOBJReader::Draw( void )
{
	m_vertArr->DrawElements( GL_TRIANGLES, 3* GetTriangleCount() );
	return IGLU_NO_ERROR;
}

int IGLUOBJReader::Draw( IGLUShaderProgram::Ptr &shader )
{
	// If the shader in question is not already enabled, enable it!
	bool wasShaderBound = shader->IsEnabled();
	if (!wasShaderBound)
		shader->Enable();

	m_vertArr->DrawElements( GL_TRIANGLES, 3* GetTriangleCount() );

	// If we started with the shader disabled, disable it again now.
	if (!wasShaderBound)
		shader->Disable();

	return IGLU_NO_ERROR;
}