/******************************************************************/
/* igluOBJReader.cpp                                              */
/* -----------------------                                        */
/*                                                                */
/* Defines a reader for OBJ model files and associated materials  */
/*                                                                */
/* Chris Wyman (10/20/2011)                                       */
/******************************************************************/

#include "iglu.h"
#include <float.h>

using namespace iglu;

namespace iglu {
	struct IGLUPointCloud::IGLUCloudHeader{
		bool resize:1;         // Was the model resized first?
		bool recenter:1;       // Was the model recentered first?
		bool useHalf:1;        // Is the data in this point cloud stored as 16-bit floats (rather than 32-bit)
		int  placeholder:29;   // Padding to fill a 32-bit chunk (reserved for later use)
		float centerX;         // Center of the point cloud
		float centerY;         
		float centerZ;
		int  dataSize;         // Number of floats/halfs in our vertex array (and the rest of the file)
	};
}

IGLUPointCloud::IGLUPointCloud( const char *filename ) :
	IGLUModel(), m_vertArr(0), 
	m_hasTexCoords(false), m_hasNormals(false), m_hasVertices(true), m_hasMatlID(false), m_hasObjectID(false),
	m_useHalf(false)
{
	// Load the cloud
	int errCode = LoadCloud( filename );
	if (errCode != IGLU_NO_ERROR)
	{
		char buf[512];
		sprintf( buf, "Error while loading point cloud '%s'!", filename );
		ErrorExit( buf, __FILE__, __FUNCTION__, __LINE__ );
	}

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	GLenum dataType = m_useHalf ? GL_HALF_FLOAT : GL_FLOAT;
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, dataType, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, dataType, m_vertStride, BUFFER_OFFSET(m_normOff));
	if (HasTexCoords())   
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, dataType, m_vertStride, BUFFER_OFFSET(m_texOff));
	if (HasMatlID())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, dataType, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	if (HasObjectID()) 
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, dataType, m_vertStride, BUFFER_OFFSET(m_objectIdOff));
}

IGLUPointCloud::IGLUPointCloud( int numPoints, unsigned short *pointData, int formatFlags ) :
	IGLUModel(), m_vertArr(0), 
	m_hasTexCoords(false), m_hasNormals(false), m_hasVertices(true), m_hasMatlID(false), m_hasObjectID(false),
	m_pointCount( numPoints ), m_useHalf(true)
{
	// Look into the format flags to find out some useful data about the point cloud.
	m_center  = (formatFlags & IGLU_PCL_WASCENTERED)   != 0;
	m_resize  = (formatFlags & IGLU_PCL_WASRESIZED)    != 0;

	// Look into the format flags to identify which types of vertex attributes are included
	m_hasVertices  = true;
	m_hasNormals   = (formatFlags & IGLU_PCL_HASNORMAL)   != 0;
	m_hasTexCoords = (formatFlags & IGLU_PCL_HASTEXCOORD) != 0;
	m_hasMatlID    = (formatFlags & IGLU_PCL_HASMATLID)   != 0;
	m_hasObjectID  = (formatFlags & IGLU_PCL_HASOBJID)    != 0;

	// Get the size of each vector component
	int sizeofElem = GetElemSz();

	// Compute the size of each vertex
	int perPointSize = ( m_hasVertices  ? 3*sizeofElem : 0 ) +
		               ( m_hasNormals   ? 3*sizeofElem : 0 ) +
					   ( m_hasTexCoords ? 2*sizeofElem : 0 ) +
					   ( m_hasMatlID    ? 1*sizeofElem : 0 ) +
					   ( m_hasObjectID  ? 1*sizeofElem : 0 );
	m_vertStride = perPointSize;

	// Compute the offsets of each vertex attribute
	m_vertOff     = 0;                                                             // The vertex always comes first
	m_normOff     = ( m_hasNormals   ? m_vertOff   + 3*sizeofElem : m_vertOff );   // The normal (if any) comes next
	m_texOff      = ( m_hasTexCoords ? m_normOff   + 3*sizeofElem : m_normOff );   // The tex coord (if any) comes next
	m_matlIdOff   = ( m_hasMatlID    ? m_texOff    + 2*sizeofElem : m_texOff );    // The material ID (if any) comes next
	m_objectIdOff = ( m_hasObjectID  ? m_matlIdOff + 1*sizeofElem : m_matlIdOff ); // The object ID (if any) comes last

	// Now that we know which data is included, and where it's all located relative to each other, go ahead and set up 
	//    our vertex array for this point cloud
	m_vertArr = new IGLUVertexArray();
	m_vertArr->SetVertexArray( numPoints * perPointSize, pointData, IGLU_STATIC|IGLU_DRAW );

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_HALF_FLOAT, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_HALF_FLOAT, m_vertStride, BUFFER_OFFSET(m_normOff));
	if (HasTexCoords())   
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_HALF_FLOAT, m_vertStride, BUFFER_OFFSET(m_texOff));
	if (HasMatlID())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, GL_HALF_FLOAT, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	if (HasObjectID()) 
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, GL_HALF_FLOAT, m_vertStride, BUFFER_OFFSET(m_objectIdOff));
}


IGLUPointCloud::IGLUPointCloud( int numPoints, float *pointData, int formatFlags ) :
	IGLUModel(), m_vertArr(0), 
	m_hasTexCoords(false), m_hasNormals(false), m_hasVertices(true), m_hasMatlID(false), m_hasObjectID(false),
	m_pointCount( numPoints ), m_useHalf(false)
{
	// Look into the format flags to find out some useful data about the point cloud.
	m_center  = (formatFlags & IGLU_PCL_WASCENTERED)   != 0;
	m_resize  = (formatFlags & IGLU_PCL_WASRESIZED)    != 0;
	m_useHalf = (formatFlags & IGLU_PCL_CONVERTTOHALF) != 0;

	// Look into the format flags to identify which types of vertex attributes are included
	m_hasVertices  = true;
	m_hasNormals   = (formatFlags & IGLU_PCL_HASNORMAL)   != 0;
	m_hasTexCoords = (formatFlags & IGLU_PCL_HASTEXCOORD) != 0;
	m_hasMatlID    = (formatFlags & IGLU_PCL_HASMATLID)   != 0;
	m_hasObjectID  = (formatFlags & IGLU_PCL_HASOBJID)    != 0;

	// Get the size of each vector component
	int sizeofElem = GetElemSz();

	// Compute the size of each vertex
	int perPointSize = ( m_hasVertices  ? 3*sizeofElem : 0 ) +
		               ( m_hasNormals   ? 3*sizeofElem : 0 ) +
					   ( m_hasTexCoords ? 2*sizeofElem : 0 ) +
					   ( m_hasMatlID    ? 1*sizeofElem : 0 ) +
					   ( m_hasObjectID  ? 1*sizeofElem : 0 );
	m_vertStride = perPointSize;

	// Compute the offsets of each vertex attribute
	m_vertOff     = 0;                                                                // The vertex always comes first
	m_normOff     = ( m_hasNormals   ? m_vertOff   + 3*sizeofElem : m_vertOff );   // The normal (if any) comes next
	m_texOff      = ( m_hasTexCoords ? m_normOff   + 3*sizeofElem : m_normOff );   // The tex coord (if any) comes next
	m_matlIdOff   = ( m_hasMatlID    ? m_texOff    + 2*sizeofElem : m_texOff );    // The material ID (if any) comes next
	m_objectIdOff = ( m_hasObjectID  ? m_matlIdOff + 1*sizeofElem : m_matlIdOff ); // The object ID (if any) comes last

	// See if we need to convert input data to half-floats...
	void *ptData = (void *)pointData;
	if (m_useHalf) // Do conversion to halfs
		ptData = ConvertToHalves( pointData, numPoints*perPointSize/sizeofElem ); 

	// Now that we know which data is included, and where it's all located relative to each other, go ahead and set up 
	//    our vertex array for this point cloud
	m_vertArr = new IGLUVertexArray();
	m_vertArr->SetVertexArray( numPoints * perPointSize, ptData, IGLU_STATIC|IGLU_DRAW );

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	GLenum dataType = m_useHalf ? GL_HALF_FLOAT : GL_FLOAT;
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, dataType, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, dataType, m_vertStride, BUFFER_OFFSET(m_normOff));
	if (HasTexCoords())   
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, dataType, m_vertStride, BUFFER_OFFSET(m_texOff));
	if (HasMatlID())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, dataType, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	if (HasObjectID()) 
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, dataType, m_vertStride, BUFFER_OFFSET(m_objectIdOff));

	// If we did conversion from float->half, we have a temporary CPU buffer hanging around. Delete.
	if (m_useHalf)
		free( ptData );
}

IGLUPointCloud::~IGLUPointCloud()
{
	// Get rid of our vertex array
	if (m_vertArr) delete m_vertArr;
}

unsigned short *IGLUPointCloud::ConvertToHalves( float *inData, int floatCount )
{
	unsigned short *ptr = (unsigned short *)malloc( floatCount * sizeof( short ) );
	for (int i=0; i<floatCount; i++)
	{
		//if ( fabs( half( inData[i] ).AsFloat() - inData[i] ) > 0.001 )
		//	printf("conversion error!  float %d\n", i);
		ptr[i] = half( inData[i] ).AsShort();
	}
	return ptr;
}

int IGLUPointCloud::LoadCloud( const char *filename )
{
	// Open the cloud file
	FILE *f = iglu::FileOpen( filename, "rb", IGLUFilePath::MODEL );
	if (!f)	return IGLU_ERROR_FILE_OPEN_FAILED;

	// Grab a file size.  This is (somewhat) non-portable code, but it's used for
	//    a sanity check for old-style point clouds that *could* have a corrupted
	//    m_useHalf bit.  For good point clouds, whether the results of this file
	//    size are non-portable should not be important
	fseek(f,0,SEEK_END);     // Find end of file
	fflush(f);               // Ensure ftell() gives us right answer
	long fileSz = ftell(f);  // Get current position
	rewind(f);               // Reset current position to top of file.

	// Load the header, copy data into the class structure
	IGLUCloudHeader header;
	fread(&header, sizeof(IGLUCloudHeader), 1, f);
	m_center   = header.recenter;
    m_resize   = header.resize;
	m_useHalf  = header.useHalf;
	m_cloudCenter = vec3( header.centerX, header.centerY, header.centerZ );

	// Sanity check.  Was m_useHalf flag corrupt in the file?  (Only happens
	//     with point clouds generated from OLD IGLU builds)
	if ( fileSz > (long)(header.dataSize * sizeof(float)) )
		m_useHalf = false;

	// Specifiy data sizes (based on half/float)
	int elemSz = GetElemSz();

	// Load the vertex data
    void *data = malloc( header.dataSize * elemSz );
    fread(data, elemSz, header.dataSize, f);

	// Store off our point count
	m_pointCount = header.dataSize/3;

	// Set up the vertex array.
	m_vertArr = new IGLUVertexArray();
	m_vertArr->SetVertexArray( header.dataSize * elemSz, data, IGLU_STATIC|IGLU_DRAW );
	m_hasVertices = true;

	// Set up striding through the array
	m_vertStride = 0;  // For now, with only vertex data...  they're densely packed
	m_vertOff    = 0;  // For now, with only vertex data... they start at byte 0!
	
	// The other offsets need to be set once we allow more data stored in the point cloud!

	// OK.  Done.
	return IGLU_NO_ERROR;
}


bool IGLUPointCloud::Save( const char *filename )
{
	//Serialize the useful class member:
    IGLUCloudHeader header;
    header.recenter = m_center;
    header.resize   = m_resize;
	header.useHalf  = m_useHalf;
	header.placeholder = 0;         // Make sure we know these values....
    header.centerX  = 0;            // We're not storing the point cloud center....
    header.centerY  = 0;
    header.centerZ  = 0;
    header.dataSize = m_vertStride * m_pointCount / GetElemSz(); 
    
	
    //Open File
	FILE *f = std::fopen( filename, "wb" );
	if (!f) return false;

    //Write header
    fwrite(&header, sizeof(IGLUCloudHeader), 1u, f);

    //Write data
	void *dataPtr = m_vertArr->MapVertexArray();
	fwrite(dataPtr, GetElemSz(), header.dataSize, f);
	m_vertArr->UnmapVertexArray();

    //Finish
    fclose(f);

	return true;
}

uint IGLUPointCloud::GetArrayBufferStride( void )
{
	return m_vertStride;
}


int IGLUPointCloud::Draw( void )
{
	m_vertArr->DrawArrays( GL_POINTS, 0, m_pointCount );
	return IGLU_NO_ERROR;
}

int IGLUPointCloud::Draw( IGLUShaderProgram::Ptr &shader )
{
	// If the shader in question is not already enabled, enable it!
	bool wasShaderBound = shader->IsEnabled();
	if (!wasShaderBound)
		shader->Enable();

	// Go ahead and draw
	m_vertArr->DrawArrays( GL_POINTS, 0, m_pointCount );

	// If we started with the shader disabled, disable it again now.
	if (!wasShaderBound)
		shader->Disable();

	return IGLU_NO_ERROR;
}