/******************************************************************/
/* igluSimpleObject.cpp                                           */
/* -----------------------                                        */
/*                                                                */
/* Defines a simple object type that developers can stuff full of */
/*    data by passing in vertex attribute arrays.                 */
/*                                                                */
/* Chris Wyman (10/24/2011)                                       */
/******************************************************************/

#include "iglu.h"
#include <float.h>

using namespace iglu;


IGLUSimpleObject::IGLUSimpleObject( void ) :
	IGLUModel(), m_vertArr(0), 
	m_numTris(0), m_numVerts(0),
	m_hasVerts(false), m_hasTexCoords(false), m_hasNorms(false), 
	m_hasMatlID(false), m_hasObjID(false)
{
	m_vertArr = new IGLUVertexArray();
}

IGLUSimpleObject::~IGLUSimpleObject()
{
	// Get rid of our vertex array
	if (m_vertArr) delete m_vertArr;
}

void IGLUSimpleObject::AddDataToArray( float *arr, int startIdx, int matlID, int objectID, float *vert, float *norm, float *tex )
{
	int i = startIdx;
	arr[i++] = float(matlID);
	arr[i++] = float(objectID);
	arr[i++] = vert[0];
	arr[i++] = vert[1];
	arr[i++] = vert[2];
	if (m_hasNorms)
	{
		arr[i++] = norm[0];
		arr[i++] = norm[1];
		arr[i++] = norm[2];
	}
	if (m_hasTexCoords)
	{
		arr[i++] = tex[0];
		arr[i++] = tex[1];
	}
}

void IGLUSimpleObject::SetGeometry( unsigned int triCount, 
								    unsigned int vertCount, 
								    unsigned int *idxBuf, 
								    float *vPos, 
		                            float *vNorm, 
									float *vTexCoord, 
									int *vMatlID, 
									int *vObjID )
{
	// Counts of the underlying data
	m_numTris  = triCount;
	m_numVerts = vertCount;

	// Identify & remember which data has been passed in
	if (vPos)      m_hasVerts     = true;
	if (vNorm)     m_hasNorms     = true;
	if (vTexCoord) m_hasTexCoords = true;
	if (vMatlID)   m_hasMatlID    = true;
	if (vObjID)    m_hasObjID     = true;

	// OK, now set up our internal IGLU storage/rendering format
	//   We'll have 1 float for a material ID
	//   We'll have 1 float for an object ID
	//   We'll have 3 floats (x,y,z) for each of the 3 verts of each triangle 
	//   We'll have 3 floats (x,y,z) for each of the 3 normals of each triangle
	//   We'll have 2 floats (u,v) for each of the 3 texture coordinates of each triangle
	uint  numComponents = 1 + 1 + 3 + (m_hasNorms ? 3 : 0) + (m_hasTexCoords ? 2 : 0);
	int bufSz           = numComponents * sizeof( float ) * m_numVerts;
	m_vertStride        = numComponents * sizeof( float );
	m_matlIdOff         = 0 * sizeof( float );
	m_objIdOff          = 1 * sizeof(float); 
	m_vertOff           = 2 * sizeof( float );
	m_normOff           = (m_hasNorms ? 5 : 0) * sizeof( float );
	m_texOff            = (m_hasTexCoords ? (m_hasNorms ? 8 : 5) : 0) * sizeof( float );

	// Create a vertex array buffer for us to fill up.
	float *tmpBuf = (float *)malloc( bufSz );

	// Fill out our vertex array with the data passed in.
	for (uint i=0; i < m_numVerts; i++ )
	{
		AddDataToArray( tmpBuf, i*numComponents, 
			            m_hasMatlID ? vMatlID[i] : 0, 
			            m_hasObjID ? vObjID[i] : 0,
						vPos+(3*i), 
						m_hasNorms ? vNorm+(3*i) : 0, 
						vTexCoord ? vTexCoord+(2*i) : 0 );
	}

	// Set our array with the interleaved data
	m_vertArr->SetVertexArray( bufSz, tmpBuf, IGLU_STATIC|IGLU_DRAW );
	m_vertArr->SetElementArray( GL_UNSIGNED_INT, m_numTris*3*sizeof(uint), idxBuf, IGLU_STATIC|IGLU_DRAW );

	// Free the temporary buffer we used to interleave the data.
	free( tmpBuf );

	// OK.  Now we've created our vertex array.  Enable the attributes so that we're sending the correct
	//    data to all our shaders.
	m_vertArr->EnableAttribute( IGLU_ATTRIB_VERTEX, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_vertOff));
	if (HasNormals())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_NORMAL, 3, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_normOff));
	if (HasTexCoords())   
		m_vertArr->EnableAttribute( IGLU_ATTRIB_TEXCOORD, 2, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_texOff));
	if (HasMatlID())  
		m_vertArr->EnableAttribute( IGLU_ATTRIB_MATL_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_matlIdOff));
	if (HasObjectID()) 
		m_vertArr->EnableAttribute( IGLU_ATTRIB_OBJECT_ID, 1, GL_FLOAT, m_vertStride, BUFFER_OFFSET(m_objIdOff));
}


int IGLUSimpleObject::Draw( void )
{
	if (!m_hasVerts)
		return IGLU_ERROR_MODEL_MISSING_DATA;

	m_vertArr->DrawElements( GL_TRIANGLES, 3*m_numTris );
	return IGLU_NO_ERROR;
}

int IGLUSimpleObject::Draw( IGLUShaderProgram::Ptr &shader )
{
	if (!m_hasVerts)
		return IGLU_ERROR_MODEL_MISSING_DATA;

	// If the shader in question is not already enabled, enable it!
	bool wasShaderBound = shader->IsEnabled();
	if (!wasShaderBound)
		shader->Enable();

	// Go ahead and draw
	m_vertArr->DrawElements( GL_TRIANGLES, 3*m_numTris );

	// If we started with the shader disabled, disable it again now.
	if (!wasShaderBound)
		shader->Disable();

	return IGLU_NO_ERROR;
}