/******************************************************************/
/* igluTexture2D.cpp                                              */
/* -----------------------                                        */
/*                                                                */
/* The file defines an image class that stores a 2D texture.      */
/*     This is very similar to the image class, but also defines  */
/*     access patterns for the texture with interpolation.        */
/*                                                                */
/* Chris Wyman (10/26/2006)                                       */
/******************************************************************/

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <GL/glew.h>
#include "iglu.h"
#include "BuiltIns/igluBuiltInTextures.h"

using namespace iglu;

extern IGLUBuiltInTexture g_igluBuiltInTex[ IGLU_BUILTIN_TEX_MAX ];

IGLUTexture2D::IGLUTexture2D( char *filename, unsigned int flags, bool initializeImmediately ) : 
	IGLUTexture(), m_imgDat(0), m_initInCtor(initializeImmediately)
{
	m_filename      = strdup( filename );
	m_texImg        = new IGLUImage( filename );
	SetTextureParameters( flags );
	m_compressTex   = (flags & IGLU_COMPRESS_TEXTURE) ? true : false;

	m_width         = m_texImg->GetWidth();
	m_height        = m_texImg->GetHeight();
	m_pixelFormat   = m_texImg->GetGLFormat();

	if (initializeImmediately) Initialize();
}

IGLUTexture2D::IGLUTexture2D( unsigned char *image, int width, int height, bool freeMemory,
							 unsigned int flags, bool initializeImmediately ) :
	IGLUTexture(), m_imgDat(0), m_initInCtor(initializeImmediately)
{
	m_filename      = strdup( "(Image from memory!)" );
	m_texImg        = new IGLUImage( image, width, height, false, freeMemory );
	SetTextureParameters( flags );
	m_compressTex   = true;

	m_width         = m_texImg->GetWidth();
	m_height        = m_texImg->GetHeight();
	m_pixelFormat   = m_texImg->GetGLFormat();

	if (initializeImmediately) Initialize();
}

IGLUTexture2D::IGLUTexture2D( int builtInTexID, unsigned int flags, bool initializeImmediately ) :
	IGLUTexture(), m_imgDat(0), m_initInCtor(initializeImmediately)
{
	if (builtInTexID < 0 || builtInTexID >= IGLU_BUILTIN_TEX_MAX)
		ErrorExit("Non-existant built-in IGLU Texture!\n", __FILE__, __FUNCTION__, __LINE__);

	m_filename      = strdup( "(Built-in IGLU texture!)" );
	m_texImg        = new IGLUImage( g_igluBuiltInTex[builtInTexID].m_pixelData, 
		                             g_igluBuiltInTex[builtInTexID].m_width, 
									 g_igluBuiltInTex[builtInTexID].m_height,
									 (g_igluBuiltInTex[builtInTexID].m_bytesPerPixel > 3) );  // has alpha?
	SetTextureParameters( flags );
	m_compressTex   = (flags & IGLU_COMPRESS_TEXTURE) ? true : false;

	m_width         = m_texImg->GetWidth();
	m_height        = m_texImg->GetHeight();
	m_pixelFormat   = m_texImg->GetGLFormat();

	if (initializeImmediately) Initialize();
}

IGLUTexture2D::IGLUTexture2D( GLenum texFormat, int width, int height, void *imgData, 
							  unsigned int flags, bool initializeImmediately ) :
	IGLUTexture(), m_pixelFormat(texFormat), m_texImg(0), m_imgDat(0), m_initInCtor(initializeImmediately)
{
	m_width         = width;
	m_height        = height;
	m_filename      = strdup( "(Image from memory!)" );
	SetTextureParameters( flags );
	m_compressTex   = false;

	if (initializeImmediately) 
	{ 
		m_imgDat = imgData; 
		Initialize();
		m_imgDat = 0;
	}
	else
	{
		int elemSz = GetFormatElementSize( texFormat );
		m_imgDat = malloc( width * height * elemSz );
		memcpy( m_imgDat, imgData, width * height * elemSz );
	}
}


IGLUTexture2D::~IGLUTexture2D()
{
	if (m_filename) free( m_filename );
	if (m_imgDat) free( m_imgDat );
}

void IGLUTexture2D::SetTextureParameters( unsigned int flags )
{
	m_minFilter     = GetMinification( flags );
	m_magFilter     = GetMagnification( flags );
	m_sWrap         = GetSWrap( flags );
	m_tWrap         = GetTWrap( flags );
	m_mipmapsNeeded = UsingMipmaps( flags );

	if (m_initialized)
	{
		glBindTexture( GL_TEXTURE_2D, m_texID );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_minFilter );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_magFilter );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, m_sWrap );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, m_tWrap );
		glBindTexture( GL_TEXTURE_2D, 0 );
	}
}

void IGLUTexture2D::Initialize( void )
{
	// Make sure to avoid doubly-calling initialize
	if (m_initialized) return;

	// Double check that we *can* initialize.
	CheckValidOpenGL( __FILE__, __FUNCTION__, __LINE__ );

	// OK, now we can set up our texture
	glBindTexture( GL_TEXTURE_2D, m_texID );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, m_minFilter );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, m_magFilter );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, m_sWrap );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, m_tWrap );

	GLenum format = CanCompressFormat( m_pixelFormat ) ? 
		GetCompressedFormat( m_pixelFormat ) : 
		m_pixelFormat;
	glTexImage2D( GL_TEXTURE_2D, 0, format,  
						   m_width, m_height, 0,
						   GetBaseInternalFormat(m_pixelFormat), GetFormatDataType(m_pixelFormat), // m_texImg->GetGLDatatype(), 
						   (m_texImg ? (void *)m_texImg->ImageData() : m_imgDat ) );
	if (m_mipmapsNeeded)
		glGenerateMipmap( GL_TEXTURE_2D );

	glBindTexture( GL_TEXTURE_2D, 0 );

	// Now that we've copied the data into texture memory, free the CPU side copy.
	if (m_texImg)
		delete m_texImg;
	if (m_imgDat && !m_initInCtor)
		free( m_imgDat );

	// Done initializing
	m_initialized = true;
}

