/******************************************************************/
/* IGLUDRand48.cpp                                                */
/* -----------------------                                        */
/*                                                                */
/* The file defines a random number generator class implementing  */
/*    drand48() and its class of functions.  This is provided in  */
/*    IGLU because drand48() is not defined on all systems (sigh) */
/*    and the IGLURandom class has limited random bits (only 16)  */
/*    which limits its usefulness for larger problems.            */
/*                                                                */
/* Chris Wyman (10/26/2006)                                       */
/******************************************************************/
#include <math.h>
#include "iglu.h"

using namespace iglu;

IGLUDRand48::IGLUDRand48( uint32_t seed )
{
	dRandAdder        = 0xb;
	dRandMultipler[0] = 0xe66d;
	dRandMultipler[1] = 0xdeec;
	dRandMultipler[2] = 0x0005;

	if (seed > 0)
	{
		dRandSeeds[0] = ushort(seed & 0xFFFF);
		dRandSeeds[1] = ushort((seed>>16) & 0xFFFF);
		dRandSeeds[2] = 0;
	}
	else
	{
		dRandSeeds[0] = 0x330e;
		dRandSeeds[1] = 0xabcd;
		dRandSeeds[2] = 0x1234;
	}
}

IGLUDRand48::IGLUDRand48( uint64_t seed )
{
	dRandAdder        = 0xb;
	dRandMultipler[0] = 0xe66d;
	dRandMultipler[1] = 0xdeec;
	dRandMultipler[2] = 0x0005;

	dRandSeeds[0] = ushort(seed & 0xFFFF);
	dRandSeeds[1] = ushort((seed>>16) & 0xFFFF);
	dRandSeeds[2] = ushort((seed>>32) & 0xFFFF);
}

void IGLUDRand48::DoRandomStep( void )
{
	ushort temp[2];

	ulong accum   = ulong(dRandMultipler[0]) * ulong(dRandSeeds[0]) + ulong(dRandAdder);
	temp[0]       = ushort(accum);
	accum       >>= sizeof(ushort) * 8;
	accum        += ulong(dRandMultipler[0]) * ulong(dRandSeeds[1]) + ulong(dRandMultipler[1]) * ulong(dRandSeeds[0]);
	temp[1]       = ushort(accum);     
	accum       >>= sizeof(ushort) * 8;
	accum        += dRandMultipler[0] * dRandSeeds[2] + dRandMultipler[1] * dRandSeeds[1] + dRandMultipler[2] * dRandSeeds[0];
	dRandSeeds[0] = temp[0];
	dRandSeeds[1] = temp[1];
	dRandSeeds[2] = ushort(accum);
}

double IGLUDRand48::dRandom( void )
{
	DoRandomStep();
	return 
		ldexp((double) dRandSeeds[0], -48) +
		ldexp((double) dRandSeeds[1], -32) +
		ldexp((double) dRandSeeds[2], -16);
}

float  IGLUDRand48::fRandom( void )
{
	DoRandomStep();
	return 
		ldexp((float) dRandSeeds[1], -32) +
		ldexp((float) dRandSeeds[2], -16);
}

unsigned int   IGLUDRand48::iRandom( void )
{
	DoRandomStep();
	return ( ulong(dRandSeeds[2]) << 16 ) + ulong(dRandSeeds[1]);
}

unsigned short IGLUDRand48::sRandom ( void )
{
	DoRandomStep();
	return dRandSeeds[2];
}

unsigned char IGLUDRand48::cRandom ( void )
{
	DoRandomStep();
	return uchar(dRandSeeds[2]>>8);
}

bool IGLUDRand48::bRandom ( void )
{
	DoRandomStep();
	return (dRandSeeds[2]>>15) > 0;  // check if the most random bit is on or off
}

vec3 IGLUDRand48::RandomHemisphereVector( void )
{
	float f1 = fRandom(), f2 = fRandom();
	return vec3( sqrt(f1)*cos(2.0f*float(IGLU_PI)*f2), 
				   sqrt(f1)*sin(2.0f*float(IGLU_PI)*f2), 
				   sqrt(1.0f - f1) );
}

vec3 IGLUDRand48::RandomSphereVector( void )
{
	float f1 = 2.0f * fRandom() - 1.0f;
	float f2 = 2.0f * float(IGLU_PI) * fRandom();
	float r = sqrt( 1.0f - f1*f1 );
	return vec3( r*cos( f2 ), r*sin( f2 ), f1 );
}

vec3 IGLUDRand48::RandomHemisphereVector( const vec3 &inDir )
{
	float f1 = fRandom(), f2 = fRandom();

	// find an orthonormal basis
	vec3 up = inDir;
	up.Normalize();
	vec3 u = ( up.X() > 0.99 || up.X() < -0.99 ? vec3::YAxis() : vec3::XAxis() );
	vec3 v = up.Cross( u );
	u = v.Cross( up );

	return (sqrt(f1)*cos(2.0f*float(IGLU_PI)*f2))*u + (sqrt(f1)*sin(2.0f*float(IGLU_PI)*f2))*v + (sqrt(1-f1))*up;

}

vec3 IGLUDRand48::RandomCosineWeightedHemisphereVector( const vec3 &inDir )
{
	float f1 = fRandom(), f2 = fRandom();
	float phi = float(IGLU_PI)*f1;
	float theta = 2.0f*float(IGLU_PI)*f2;

	// find an orthonormal basis
	vec3 up = inDir;
	up.Normalize();
	vec3 u = ( up.X() > 0.99 || up.X() < -0.99 ? vec3::YAxis() : vec3::XAxis() );
	vec3 v = up.Cross( u );
	u = v.Cross( up );

	return sin(phi)*cos(theta)*u + sin(phi)*sin(theta)*v + cos(phi)*up;
}


